/*
 * base64.c -- main test file for base64
 *
 * Copyright (c) 2013-2022 David Demelier <markand@malikania.fr>
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include <errno.h>

#include <rexo.h>

#include <base64.h>

RX_TEST_CASE(general, b64_lookup)
{
	RX_INT_REQUIRE_EQUAL(b64_lookup(0), 'A');
	RX_INT_REQUIRE_EQUAL(b64_lookup(1), 'B');
	RX_INT_REQUIRE_EQUAL(b64_lookup(2), 'C');
	RX_INT_REQUIRE_EQUAL(b64_lookup(3), 'D');
	RX_INT_REQUIRE_EQUAL(b64_lookup(4), 'E');
	RX_INT_REQUIRE_EQUAL(b64_lookup(5), 'F');
	RX_INT_REQUIRE_EQUAL(b64_lookup(6), 'G');
	RX_INT_REQUIRE_EQUAL(b64_lookup(7), 'H');
	RX_INT_REQUIRE_EQUAL(b64_lookup(8), 'I');
	RX_INT_REQUIRE_EQUAL(b64_lookup(9), 'J');
	RX_INT_REQUIRE_EQUAL(b64_lookup(10), 'K');
	RX_INT_REQUIRE_EQUAL(b64_lookup(11), 'L');
	RX_INT_REQUIRE_EQUAL(b64_lookup(12), 'M');
	RX_INT_REQUIRE_EQUAL(b64_lookup(13), 'N');
	RX_INT_REQUIRE_EQUAL(b64_lookup(14), 'O');
	RX_INT_REQUIRE_EQUAL(b64_lookup(15), 'P');
	RX_INT_REQUIRE_EQUAL(b64_lookup(16), 'Q');
	RX_INT_REQUIRE_EQUAL(b64_lookup(17), 'R');
	RX_INT_REQUIRE_EQUAL(b64_lookup(18), 'S');
	RX_INT_REQUIRE_EQUAL(b64_lookup(19), 'T');
	RX_INT_REQUIRE_EQUAL(b64_lookup(20), 'U');
	RX_INT_REQUIRE_EQUAL(b64_lookup(21), 'V');
	RX_INT_REQUIRE_EQUAL(b64_lookup(22), 'W');
	RX_INT_REQUIRE_EQUAL(b64_lookup(23), 'X');
	RX_INT_REQUIRE_EQUAL(b64_lookup(24), 'Y');
	RX_INT_REQUIRE_EQUAL(b64_lookup(25), 'Z');
	RX_INT_REQUIRE_EQUAL(b64_lookup(26), 'a');
	RX_INT_REQUIRE_EQUAL(b64_lookup(27), 'b');
	RX_INT_REQUIRE_EQUAL(b64_lookup(28), 'c');
	RX_INT_REQUIRE_EQUAL(b64_lookup(29), 'd');
	RX_INT_REQUIRE_EQUAL(b64_lookup(30), 'e');
	RX_INT_REQUIRE_EQUAL(b64_lookup(31), 'f');
	RX_INT_REQUIRE_EQUAL(b64_lookup(32), 'g');
	RX_INT_REQUIRE_EQUAL(b64_lookup(33), 'h');
	RX_INT_REQUIRE_EQUAL(b64_lookup(34), 'i');
	RX_INT_REQUIRE_EQUAL(b64_lookup(35), 'j');
	RX_INT_REQUIRE_EQUAL(b64_lookup(36), 'k');
	RX_INT_REQUIRE_EQUAL(b64_lookup(37), 'l');
	RX_INT_REQUIRE_EQUAL(b64_lookup(38), 'm');
	RX_INT_REQUIRE_EQUAL(b64_lookup(39), 'n');
	RX_INT_REQUIRE_EQUAL(b64_lookup(40), 'o');
	RX_INT_REQUIRE_EQUAL(b64_lookup(41), 'p');
	RX_INT_REQUIRE_EQUAL(b64_lookup(42), 'q');
	RX_INT_REQUIRE_EQUAL(b64_lookup(43), 'r');
	RX_INT_REQUIRE_EQUAL(b64_lookup(44), 's');
	RX_INT_REQUIRE_EQUAL(b64_lookup(45), 't');
	RX_INT_REQUIRE_EQUAL(b64_lookup(46), 'u');
	RX_INT_REQUIRE_EQUAL(b64_lookup(47), 'v');
	RX_INT_REQUIRE_EQUAL(b64_lookup(48), 'w');
	RX_INT_REQUIRE_EQUAL(b64_lookup(49), 'x');
	RX_INT_REQUIRE_EQUAL(b64_lookup(50), 'y');
	RX_INT_REQUIRE_EQUAL(b64_lookup(51), 'z');
	RX_INT_REQUIRE_EQUAL(b64_lookup(52), '0');
	RX_INT_REQUIRE_EQUAL(b64_lookup(53), '1');
	RX_INT_REQUIRE_EQUAL(b64_lookup(54), '2');
	RX_INT_REQUIRE_EQUAL(b64_lookup(55), '3');
	RX_INT_REQUIRE_EQUAL(b64_lookup(56), '4');
	RX_INT_REQUIRE_EQUAL(b64_lookup(57), '5');
	RX_INT_REQUIRE_EQUAL(b64_lookup(58), '6');
	RX_INT_REQUIRE_EQUAL(b64_lookup(59), '7');
	RX_INT_REQUIRE_EQUAL(b64_lookup(60), '8');
	RX_INT_REQUIRE_EQUAL(b64_lookup(61), '9');
	RX_INT_REQUIRE_EQUAL(b64_lookup(62), '+');
	RX_INT_REQUIRE_EQUAL(b64_lookup(63), '/');
}

RX_TEST_CASE(general, b64_rlookup)
{
	RX_INT_REQUIRE_EQUAL(b64_rlookup('A'), 0);
	RX_INT_REQUIRE_EQUAL(b64_rlookup('B'), 1);
	RX_INT_REQUIRE_EQUAL(b64_rlookup('C'), 2);
	RX_INT_REQUIRE_EQUAL(b64_rlookup('D'), 3);
	RX_INT_REQUIRE_EQUAL(b64_rlookup('E'), 4);
	RX_INT_REQUIRE_EQUAL(b64_rlookup('F'), 5);
	RX_INT_REQUIRE_EQUAL(b64_rlookup('G'), 6);
	RX_INT_REQUIRE_EQUAL(b64_rlookup('H'), 7);
	RX_INT_REQUIRE_EQUAL(b64_rlookup('I'), 8);
	RX_INT_REQUIRE_EQUAL(b64_rlookup('J'), 9);
	RX_INT_REQUIRE_EQUAL(b64_rlookup('K'), 10);
	RX_INT_REQUIRE_EQUAL(b64_rlookup('L'), 11);
	RX_INT_REQUIRE_EQUAL(b64_rlookup('M'), 12);
	RX_INT_REQUIRE_EQUAL(b64_rlookup('N'), 13);
	RX_INT_REQUIRE_EQUAL(b64_rlookup('O'), 14);
	RX_INT_REQUIRE_EQUAL(b64_rlookup('P'), 15);
	RX_INT_REQUIRE_EQUAL(b64_rlookup('Q'), 16);
	RX_INT_REQUIRE_EQUAL(b64_rlookup('R'), 17);
	RX_INT_REQUIRE_EQUAL(b64_rlookup('S'), 18);
	RX_INT_REQUIRE_EQUAL(b64_rlookup('T'), 19);
	RX_INT_REQUIRE_EQUAL(b64_rlookup('U'), 20);
	RX_INT_REQUIRE_EQUAL(b64_rlookup('V'), 21);
	RX_INT_REQUIRE_EQUAL(b64_rlookup('W'), 22);
	RX_INT_REQUIRE_EQUAL(b64_rlookup('X'), 23);
	RX_INT_REQUIRE_EQUAL(b64_rlookup('Y'), 24);
	RX_INT_REQUIRE_EQUAL(b64_rlookup('Z'), 25);
	RX_INT_REQUIRE_EQUAL(b64_rlookup('a'), 26);
	RX_INT_REQUIRE_EQUAL(b64_rlookup('b'), 27);
	RX_INT_REQUIRE_EQUAL(b64_rlookup('c'), 28);
	RX_INT_REQUIRE_EQUAL(b64_rlookup('d'), 29);
	RX_INT_REQUIRE_EQUAL(b64_rlookup('e'), 30);
	RX_INT_REQUIRE_EQUAL(b64_rlookup('f'), 31);
	RX_INT_REQUIRE_EQUAL(b64_rlookup('g'), 32);
	RX_INT_REQUIRE_EQUAL(b64_rlookup('h'), 33);
	RX_INT_REQUIRE_EQUAL(b64_rlookup('i'), 34);
	RX_INT_REQUIRE_EQUAL(b64_rlookup('j'), 35);
	RX_INT_REQUIRE_EQUAL(b64_rlookup('k'), 36);
	RX_INT_REQUIRE_EQUAL(b64_rlookup('l'), 37);
	RX_INT_REQUIRE_EQUAL(b64_rlookup('m'), 38);
	RX_INT_REQUIRE_EQUAL(b64_rlookup('n'), 39);
	RX_INT_REQUIRE_EQUAL(b64_rlookup('o'), 40);
	RX_INT_REQUIRE_EQUAL(b64_rlookup('p'), 41);
	RX_INT_REQUIRE_EQUAL(b64_rlookup('q'), 42);
	RX_INT_REQUIRE_EQUAL(b64_rlookup('r'), 43);
	RX_INT_REQUIRE_EQUAL(b64_rlookup('s'), 44);
	RX_INT_REQUIRE_EQUAL(b64_rlookup('t'), 45);
	RX_INT_REQUIRE_EQUAL(b64_rlookup('u'), 46);
	RX_INT_REQUIRE_EQUAL(b64_rlookup('v'), 47);
	RX_INT_REQUIRE_EQUAL(b64_rlookup('w'), 48);
	RX_INT_REQUIRE_EQUAL(b64_rlookup('x'), 49);
	RX_INT_REQUIRE_EQUAL(b64_rlookup('y'), 50);
	RX_INT_REQUIRE_EQUAL(b64_rlookup('z'), 51);
	RX_INT_REQUIRE_EQUAL(b64_rlookup('0'), 52);
	RX_INT_REQUIRE_EQUAL(b64_rlookup('1'), 53);
	RX_INT_REQUIRE_EQUAL(b64_rlookup('2'), 54);
	RX_INT_REQUIRE_EQUAL(b64_rlookup('3'), 55);
	RX_INT_REQUIRE_EQUAL(b64_rlookup('4'), 56);
	RX_INT_REQUIRE_EQUAL(b64_rlookup('5'), 57);
	RX_INT_REQUIRE_EQUAL(b64_rlookup('6'), 58);
	RX_INT_REQUIRE_EQUAL(b64_rlookup('7'), 59);
	RX_INT_REQUIRE_EQUAL(b64_rlookup('8'), 60);
	RX_INT_REQUIRE_EQUAL(b64_rlookup('9'), 61);
	RX_INT_REQUIRE_EQUAL(b64_rlookup('+'), 62);
	RX_INT_REQUIRE_EQUAL(b64_rlookup('/'), 63);
}

RX_TEST_CASE(encode, simple)
{
	char buffer[BUFSIZ];
	size_t r;

	r = b64_encode("a", -1, buffer, sizeof (buffer));
	RX_UINT_REQUIRE_EQUAL(r, 4U);
	RX_STR_REQUIRE_EQUAL(buffer, "YQ==");
	r = b64_encode("ab", -1, buffer, sizeof (buffer));
	RX_UINT_REQUIRE_EQUAL(r, 4U);
	RX_STR_REQUIRE_EQUAL(buffer, "YWI=");
	r = b64_encode("abc", -1, buffer, sizeof (buffer));
	RX_UINT_REQUIRE_EQUAL(r, 4U);
	RX_STR_REQUIRE_EQUAL(buffer, "YWJj");
	r = b64_encode("hello", -1, buffer, sizeof (buffer));
	RX_UINT_REQUIRE_EQUAL(r, 8U);
	RX_STR_REQUIRE_EQUAL(buffer, "aGVsbG8=");
	r = b64_encode("this is a long sentence", -1, buffer, sizeof (buffer));
	RX_UINT_REQUIRE_EQUAL(r, 32U);
	RX_STR_REQUIRE_EQUAL(buffer, "dGhpcyBpcyBhIGxvbmcgc2VudGVuY2U=");
}

RX_TEST_CASE(encode, binary)
{
	char buffer[BUFSIZ];
	size_t r;

	r = b64_encode("a\0bc", 4, buffer, sizeof (buffer));
	RX_UINT_REQUIRE_EQUAL(r, 8U);
	RX_STR_REQUIRE_EQUAL(buffer, "YQBiYw==");
}

RX_TEST_CASE(encode, toosmall)
{
	char buffer[10];
	size_t r;

	r = b64_encode("hello world", -1, buffer, sizeof (buffer));
	RX_UINT_REQUIRE_EQUAL(r, (size_t)-1);
	RX_INT_REQUIRE_EQUAL(errno, ERANGE);
}

RX_TEST_CASE(encode, length)
{
	RX_UINT_REQUIRE_EQUAL(B64_ENCODE_LENGTH(1), 4U);
	RX_UINT_REQUIRE_EQUAL(B64_ENCODE_LENGTH(2), 4U);
	RX_UINT_REQUIRE_EQUAL(B64_ENCODE_LENGTH(3), 8U);
	RX_UINT_REQUIRE_EQUAL(B64_ENCODE_LENGTH(4), 8U);
	RX_UINT_REQUIRE_EQUAL(B64_ENCODE_LENGTH(5), 8U);
	RX_UINT_REQUIRE_EQUAL(B64_ENCODE_LENGTH(6), 12U);
	RX_UINT_REQUIRE_EQUAL(B64_ENCODE_LENGTH(7), 12U);
	RX_UINT_REQUIRE_EQUAL(B64_ENCODE_LENGTH(8), 12U);
}

RX_TEST_CASE(decode, length)
{
	RX_UINT_REQUIRE_EQUAL(B64_DECODE_LENGTH(4), 3U);
	RX_UINT_REQUIRE_EQUAL(B64_DECODE_LENGTH(8), 6U);
	RX_UINT_REQUIRE_EQUAL(B64_DECODE_LENGTH(12), 9U);
	RX_UINT_REQUIRE_EQUAL(B64_DECODE_LENGTH(16), 12U);
	RX_UINT_REQUIRE_EQUAL(B64_DECODE_LENGTH(20), 15U);
}

RX_TEST_CASE(decode, simple)
{
	static const struct {
		const char *entry;
		const char *exp;
		size_t expsz;
	} examples[] = {
		{ "YQ==", "a", 1U },
		{ "YWI=", "ab", 2U },
		{ "YWJj", "abc", 3U },
		{ "aGVsbG8=", "hello", 5U },
		{ "dGhpcyBpcyBhIGxvbmcgc2VudGVuY2U=", "this is a long sentence", 23U },
		{ "V2VsY29tZSB0byBvdXIgc2VydmVyIGR1ZGU=", "Welcome to our server dude", 26U },
		{ NULL, NULL, 0 }
	};

	char buffer[BUFSIZ];
	size_t r;

	for (size_t i = 0; examples[i].entry; ++i) {
		r = b64_decode(examples[i].entry, -1, buffer, sizeof (buffer));
		RX_UINT_REQUIRE_EQUAL(r, examples[i].expsz);
		buffer[r] = '\0';
		RX_STR_REQUIRE_EQUAL(buffer, examples[i].exp);
	}
}

RX_TEST_CASE(decode, binary)
{
	static const struct {
		const char *entry;
		size_t expsz;
	} examples[] = {
		{ "4oARcAAAAgoogyHe", 12 },
		{ "ERADdw==", 4 },
		{ "4oARkaUCAGAbAtfN", 12 },
		{ NULL, 0 }
	};

	char bin[BUFSIZ], b64[BUFSIZ];
	size_t r;

	for (size_t i = 0; examples[i].entry; ++i) {
		/* Compare by decoding and coding back again. */
		r = b64_decode(examples[i].entry, -1, bin, sizeof (bin));
		RX_UINT_REQUIRE_EQUAL(r, examples[i].expsz);
		r = b64_encode(bin, r, b64, sizeof (b64));
		RX_UINT_REQUIRE_EQUAL(r, strlen(examples[i].entry));
		RX_STR_REQUIRE_EQUAL(b64, examples[i].entry);
	}
}

RX_TEST_CASE(decode, sized)
{
	char buffer[BUFSIZ];
	size_t r;

	r = b64_decode("YQ==     ", 4, buffer, sizeof (buffer));
	RX_UINT_REQUIRE_EQUAL(r, 1U);
	buffer[r] = '\0';
	RX_STR_REQUIRE_EQUAL(buffer, "a");
}

RX_TEST_CASE(decode, invalid)
{
	char buffer[BUFSIZ];
	size_t r;

	r = b64_decode("YW=", -1, buffer, sizeof (buffer));
	RX_UINT_REQUIRE_EQUAL(r, (size_t)-1);
	RX_INT_REQUIRE_EQUAL(errno, EILSEQ);
	r = b64_decode("?!", -1, buffer, sizeof (buffer));
	RX_UINT_REQUIRE_EQUAL(r, (size_t)-1);
	RX_INT_REQUIRE_EQUAL(errno, EILSEQ);
	r = b64_decode("=ABC", -1, buffer, sizeof (buffer));
	RX_UINT_REQUIRE_EQUAL(r, (size_t)-1);
	RX_INT_REQUIRE_EQUAL(errno, EILSEQ);
	r = b64_decode("A=BC", -1, buffer, sizeof (buffer));
	RX_UINT_REQUIRE_EQUAL(r, (size_t)-1);
	RX_INT_REQUIRE_EQUAL(errno, EILSEQ);
	r = b64_decode("==BC", -1, buffer, sizeof (buffer));
	RX_UINT_REQUIRE_EQUAL(r, (size_t)-1);
	RX_INT_REQUIRE_EQUAL(errno, EILSEQ);
	r = b64_decode("AB=C", -1, buffer, sizeof (buffer));
	RX_UINT_REQUIRE_EQUAL(r, (size_t)-1);
	RX_INT_REQUIRE_EQUAL(errno, EILSEQ);
}

RX_TEST_CASE(decode, toosmall)
{
	char buf[100];
	size_t r;

	/* Requires 1. */
	for (size_t i = 0; i < 1; ++i) {
		r = b64_decode("MQ==", -1, buf, i);
		RX_UINT_REQUIRE_EQUAL(r, (size_t)-1);
		RX_INT_REQUIRE_EQUAL(errno, ERANGE);
	}

	/* Requires 2. */
	for (size_t i = 0; i < 2; ++i) {
		r = b64_decode("MTI=", -1, buf, i);
		RX_UINT_REQUIRE_EQUAL(r, (size_t)-1);
		RX_INT_REQUIRE_EQUAL(errno, ERANGE);
	}

	/* Requires 3. */
	for (size_t i = 0; i < 3; ++i) {
		r = b64_decode("MTIz", -1, buf, i);
		RX_UINT_REQUIRE_EQUAL(r, (size_t)-1);
		RX_INT_REQUIRE_EQUAL(errno, ERANGE);
	}

	/* Requires 4. */
	for (size_t i = 0; i < 4; ++i) {
		r = b64_decode("YWJjZA==", -1, buf, i);
		RX_UINT_REQUIRE_EQUAL(r, (size_t)-1);
		RX_INT_REQUIRE_EQUAL(errno, ERANGE);
	}

	/* Requires 5. */
	for (size_t i = 0; i < 5; ++i) {
		r = b64_decode("YWJjZGU=", -1, buf, i);
		RX_UINT_REQUIRE_EQUAL(r, (size_t)-1);
		RX_INT_REQUIRE_EQUAL(errno, ERANGE);
	}

	/* Requires 6. */
	for (size_t i = 0; i < 6; ++i) {
		r = b64_decode("YWJjZGVm", -1, buf, i);
		RX_UINT_REQUIRE_EQUAL(r, (size_t)-1);
		RX_INT_REQUIRE_EQUAL(errno, ERANGE);
	}

	/* Requires 7. */
	for (size_t i = 0; i < 7; ++i) {
		r = b64_decode("YWJjZGVmZw==", -1, buf, i);
		RX_UINT_REQUIRE_EQUAL(r, (size_t)-1);
		RX_INT_REQUIRE_EQUAL(errno, ERANGE);
	}

	/* Requires 8. */
	for (size_t i = 0; i < 8; ++i) {
		r = b64_decode("YWJjZGVmZ2g=", -1, buf, i);
		RX_UINT_REQUIRE_EQUAL(r, (size_t)-1);
		RX_INT_REQUIRE_EQUAL(errno, ERANGE);
	}

	/* Requires 9. */
	for (size_t i = 0; i < 9; ++i) {
		r = b64_decode("MTIzNDU2Nzg5", -1, buf, i);
		RX_UINT_REQUIRE_EQUAL(r, (size_t)-1);
		RX_INT_REQUIRE_EQUAL(errno, ERANGE);
	}

	/* Requires 10. */
	for (size_t i = 0; i < 10; ++i) {
		r = b64_decode("MTIzNDU2Nzg5eA==", -1, buf, i);
		RX_UINT_REQUIRE_EQUAL(r, (size_t)-1);
		RX_INT_REQUIRE_EQUAL(errno, ERANGE);
	}
}

int
main(int argc, char **argv)
{
	return rx_main(0, NULL, argc, (const char * const *)argv) == RX_SUCCESS ? 0 : 1;
}
